//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 1.0, 2007/03/08
//  Generated by Andy Kiser
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
// Originally authored by Toradex AG. This work is now available under the terms and conditions
// of the Creative Commons License 'Attribution CC BY', details of which can be found here:
// http://creativecommons.org/licenses/by/3.0/
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,		// report mode
	0x00,		// LED mode
	0x02		// Timing register, specific for TSL2561
				// Bit 7	synchronized sampling (to be implemented by Toradex Firmware using INT)
				// Bit 6..5	RSVD, always 0
				// Bit 4	GAIN (0=1x, 1=16x)
				// Bit 3	Manual Integrator control (0=off, 1=on)
				// Bit 2	RSVE, always 0
				// Bit 1..0 Integration time (0=13.7ms, 1=101ms, 2=402ms, 3=MANUAL)	
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	500,		// report rate [ms]
	500,		// sample rate [ms]
	50000		// sample rate [us]	(not used for this sensor)
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Illuminance Sensor",	// user device name
	"Frame Number",			// user name Channel 0
	"Illuminance"			// user name Channel 1
};


//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;

//----------------------------------------
// constant lookup-table for conversion ADC->physical value
//----------------------------------------


//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	eReset,
	eIdle,
	eReadADC0,
	eReadADC1,
	eCalculating
} tFSM_Measure;

//----------------------------------------
// variables

tFSM_Measure	FSM;
WORD			Countdown_1ms;

BOOL			NewSample = FALSE;

//---------------------------------------------------
// Global defines
//
#define SLAVEADDR 0x29

//---------------------------------------------------
// Gain constants (for parameter passing)
//
#define GAIN_1x 0
#define GAIN_16x 1

//---------------------------------------------------
// Integration Time constants (for parameter passing)
//
#define TIM_13_7ms 0
#define TIM_100ms 1
#define TIM_402ms 2
#define TIM_MANUAL 3

// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM2_2, DM2_1, DM2_0;

//----------------------------------------
// function prototypes
//
static BYTE TSL2561_TxRx (BYTE TxData, BYTE size);
static void TSL2561_WriteByte(BYTE Addr, BYTE Command, BYTE Data);
static WORD TSL2561_ReadWord(BYTE Addr, BYTE Command);
static BYTE TSL2561_ReadByte(BYTE Addr, BYTE Command);
static void TSL2561_WriteByte(BYTE Addr, BYTE Command, BYTE Data);
static WORD CalculateLux (BYTE iGain, BYTE tInt, WORD ch0, WORD ch1);

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	FSM = eReset;
	// make sure INT is high
	INT_DataShadow |= INT_MASK;
	INT_Data_ADDR = INT_DataShadow;
	
	// clock I2C bus to reset the TSL2561 to a defined state (if it is powered on)
	TSL2561_TxRx(0xFF,8);
	TSL2561_TxRx(0xFF,8);
	
	TSL2561_WriteByte(SLAVEADDR, 0xC0, 0x03); // power on
	TSL2561_WriteByte(SLAVEADDR, 0xC1, 0x012); // switch to high gain (16x)
}

//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;
	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;
		
	// send reset sequence to stop current measurement
	TSL2561_WriteByte(SLAVEADDR, 0xC0, 0x00); // power off

	// set all I/Os to/from the sensor to High-Z Analog (=110), 
	PRT0DM2  = 0xFF;
	PRT0DM1  = 0xFF;
	PRT0DM0  = 0x00;
	PRT2DM2  = 0xFF;
	PRT2DM1  = 0xFF;
	PRT2DM0  = 0x00;
}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	// restore drivemodes for GPIOs to/from the sensor
	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;

	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;

	FSM = eReset;
	TSL2561_WriteByte(SLAVEADDR, 0xC0, 0x03); // power on
}


//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	// currently no special feature reports are handled
	return 0;
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	// Oak-Lux does not have specific parameters
}


//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
void Sensor_FSM_Measure() 
{
	static WORD Ch0; // visible radiation
	static WORD Ch1; // infrared radiation
	WORD        Lux;
	static BYTE TIMING_Readback = 0x02; // copy of TSL2561's TIMING register
	BYTE        TIMING_ToBe;
	BYTE		Gain, IntegrationTime;

	// decrement millisecond counter
	if (Countdown_1ms != 0)
		Countdown_1ms--;

	// process actual state machine
	switch (FSM)
	{
		case eReset:
			Countdown_1ms = 450; // 450ms delay until first measurement finished
			
			TIMING_Readback = TSL2561_ReadByte(SLAVEADDR, 0xC1);
			
			FSM = eIdle;
			break;
			
		case eIdle: 
			Get_Param_8(PAR8_TIMING, VOLATILE, &TIMING_ToBe);
			TIMING_ToBe &= 0x1B; // all other bits are reserved and read 0
			if (TIMING_ToBe != TIMING_Readback)
			{
				TSL2561_WriteByte(SLAVEADDR, 0xC1, TIMING_ToBe);
				TIMING_Readback = TSL2561_ReadByte(SLAVEADDR, 0xC1);
			}
			if (Countdown_1ms == 0)
			{
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);;
				FSM = eReadADC0;					  
			}
			break;			
			
		case eReadADC0:
			Ch0 = TSL2561_ReadWord(SLAVEADDR, 0xAC);
			FSM = eReadADC1;
			break;			

		case eReadADC1:
			Ch1 = TSL2561_ReadWord(SLAVEADDR, 0xAE);
			FSM = eCalculating;
			break;			
		
		case eCalculating:
			Gain = (TIMING_Readback &0x10) >> 4;	// Bit 4 of the register is the gain
			IntegrationTime = TIMING_Readback & 0x03; // Bit 1..0 are the integration time
			Lux = CalculateLux(Gain, IntegrationTime, Ch0, Ch1);

			// fill measured data into IN REPORT
			Report.Ch1_LSB = Lux & 0xFF;
			Report.Ch1_MSB = Lux / 0x100;

			NewSample=TRUE;
			
			FSM = eIdle;
			break;			
	}
}

//-----------------------------------------------------------------------------
// Transmit Start Condition
//-----------------------------------------------------------------------------
static void TSL2561_TxStart (void)
{
	// make sure SDA is high
	SDA_DataShadow |= SDA_MASK;
	SDA_Data_ADDR = SDA_DataShadow;
	
	// make sure SCL is high
	SCL_DataShadow |= SCL_MASK;
	SCL_Data_ADDR = SCL_DataShadow;
	
	// send Start Condition: negative edge on SDA while SCL is high
	SDA_DataShadow &= ~SDA_MASK;
	SDA_Data_ADDR = SDA_DataShadow;
}

//-----------------------------------------------------------------------------
// Transmit Stop Condition
//-----------------------------------------------------------------------------
void TSL2561_TxStop (void)
{
	BYTE i;
	
	// make sure SDA is low
	SDA_DataShadow &= ~SDA_MASK;
	SDA_Data_ADDR = SDA_DataShadow;
	
	// make sure SCL is high
	SCL_DataShadow |= SCL_MASK;
	SCL_Data_ADDR = SCL_DataShadow;
	
	// wait 0.5 cycles
	for(i=0; i<1; i++);
	
	// send Start Condition: positive edge on SDA while SCL is high
	SDA_DataShadow |= SDA_MASK;
	SDA_Data_ADDR = SDA_DataShadow;

	// wait 0.5 cycles
	for(i=0; i<1; i++);
}

//-----------------------------------------------------------------------------
// Transmit and Receive <size> bits to the sensor / from the sensor
// - Data bits 0..<size-1> are transmitted, MSB first
// - <size> must be in the range 0..8
// return value: received bits 
//-----------------------------------------------------------------------------
static BYTE TSL2561_TxRx (BYTE TxData, BYTE size)
{
	BYTE i;
	BYTE Rx=0;
	
	// shift TxData to the left end of the byte
	for (i=size; i<8; i++)
		TxData <<=1;
	
	// make sure SCL is low
	SCL_DataShadow &= ~SCL_MASK;
	SCL_Data_ADDR = SCL_DataShadow;
	
	while (size != 0)
	{
		// set SDA 
		if ( (TxData & 0x80) == 0 )
			SDA_DataShadow &= ~SDA_MASK;
		else
			SDA_DataShadow |=  SDA_MASK;
		SDA_Data_ADDR = SDA_DataShadow;
		
		// wait 0.5 cycles (use time to do some processing)
		TxData <<= 1;
		Rx <<= 1;
		
		// generate positive clock edge
		SCL_DataShadow |= SCL_MASK;
		SCL_Data_ADDR = SCL_DataShadow;
		
		// wait 0.5 cycles (use time to do some processing)
		// read Rx data bit
		if ( (SDA_Data_ADDR & SDA_MASK) != 0)
			Rx++;
		size--;
		
		// generate negative clock edge
		SCL_DataShadow &= ~SCL_MASK;
		SCL_Data_ADDR = SCL_DataShadow;
		
	}	
	
	return Rx;
}
//-----------------------------------------------------------------------------
// Receive Acknowledge bit
// return TRUE if Acknowledge received
//-----------------------------------------------------------------------------
BOOL TSL2561_RxAck()
{
	BYTE Ack;

	// make sure SCL is low
	SCL_DataShadow &= ~SCL_MASK;
	SCL_Data_ADDR = SCL_DataShadow;

	// make sure SDA is high
	SDA_DataShadow |= SDA_MASK;
	SDA_Data_ADDR = SDA_DataShadow;

	Ack = TSL2561_TxRx(0x01, 1);
	return (Ack == 0);
}


//-----------------------------------------------------------------------------
// Transmit one byte using the "SendByte" Protocol
//-----------------------------------------------------------------------------
static void TSL2561_WriteByte(BYTE Addr, BYTE Command, BYTE Data)
{
	BOOL Ack;
	
	// Send Start Condition
	TSL2561_TxStart();
	
	// Transmit Address and R/nW Bit
	TSL2561_TxRx (Addr <<1, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();
	
	// Transmit Command Byte
	TSL2561_TxRx (Command, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();

	// Transmit Data Byte
	TSL2561_TxRx (Data, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();

	// Send Stop Condition
	TSL2561_TxStop();
}
//-----------------------------------------------------------------------------
// Receive one byte using the "ReceiveByte" Protocol
// command must be (0xC0 + Register Address)
//-----------------------------------------------------------------------------
static BYTE TSL2561_ReadByte(BYTE Addr, BYTE Command)
{
	BOOL Ack;
	BYTE Data;
	
	// Send Start Condition
	TSL2561_TxStart();
	
	// Transmit Address and R/nW Bit
	Addr <<= 1;
	TSL2561_TxRx (Addr, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();
	
	// Transmit Command Byte
	TSL2561_TxRx (Command, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();

	// Send Start Condition
	TSL2561_TxStart();
	
	// Transmit Address and R/nW Bit
	TSL2561_TxRx (++Addr, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();

	// Receive Data Byte
	Data = TSL2561_TxRx (0xFF, 8);

	// Send NAK to finish transfer
	TSL2561_TxRx(0xFF,1);  

	// Send Stop Condition
	TSL2561_TxStop();

	return Data;
}

//-----------------------------------------------------------------------------
// Receive 16 bit word using the "ReceiveWord" Protocol(LittleEndian, LSB first)
// command must be (0xE0 + Register Address)
//-----------------------------------------------------------------------------
static WORD TSL2561_ReadWord(BYTE Addr, BYTE Command)
{
	BOOL Ack;
	union uBigEndian Data;

	
	// Send Start Condition
	TSL2561_TxStart();
	
	// Transmit Address and R/nW Bit
	Addr <<= 1;
	TSL2561_TxRx (Addr, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();
	
	// Transmit Command Byte
	TSL2561_TxRx (Command, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();

	// Send Start Condition
	TSL2561_TxStart();
	
	// Transmit Address and R/nW Bit
	TSL2561_TxRx (++Addr, 8);
	// Read Acknowledge
	Ack = TSL2561_RxAck();

	// Receive Low Data Byte
	Data.BE.Lo = TSL2561_TxRx (0xFF, 8);

	// Send ACK
	TSL2561_TxRx(0x00,1);  
	
	// Receive High Data Byte
	Data.BE.Hi = TSL2561_TxRx (0xFF, 8);
	
	// Send NAK to finish transfer
	TSL2561_TxRx(0xFF,1);  

	// Send Stop Condition
	TSL2561_TxStop();

	return Data.w;
}


//---------------------------------------------------
// Scale factors
//
#define LUX_SCALE 14 // scale by 2^14
#define RATIO_SCALE 9 // scale ratio by 2^9

//---------------------------------------------------
// Integration time scaling factors
//
#define CH_SCALE 10 // scale channel values by 2^10
#define CHSCALE_TINT0 0x7517 // 322/11 * 2^CH_SCALE
#define CHSCALE_TINT1 0x0fe7 // 322/81 * 2^CH_SCALE

//---------------------------------------------------
// T Package coefficients
//---------------------------------------------------
// For Ch1/Ch0=0.00 to 0.50
// 		Lux/Ch0=0.0304-0.062*((Ch1/Ch0)^1.4)
// 		piecewise approximation
// 			For Ch1/Ch0=0.00 to 0.125:
// 				Lux/Ch0=0.0304-0.0272*(Ch1/Ch0)
//
// 			For Ch1/Ch0=0.125 to 0.250:
// 				Lux/Ch0=0.0325-0.0440*(Ch1/Ch0)
//
// 			For Ch1/Ch0=0.250 to 0.375:
// 				Lux/Ch0=0.0351-0.0544*(Ch1/Ch0)
//
// 			For Ch1/Ch0=0.375 to 0.50:
// 				Lux/Ch0=0.0381-0.0624*(Ch1/Ch0)
//
// For Ch1/Ch0=0.50 to 0.61:
// 		Lux/Ch0=0.0224-0.031*(Ch1/Ch0)
//
// For Ch1/Ch0=0.61 to 0.80:
// 		Lux/Ch0=0.0128-0.0153*(Ch1/Ch0)
//
// For Ch1/Ch0=0.80 to 1.30:
// 		Lux/Ch0=0.00146-0.00112*(Ch1/Ch0)
//
// For Ch1/Ch0>1.3:
// 		Lux/Ch0=0
//---------------------------------------------------
#define K1T 0x0040 // 0.125 * 2^RATIO_SCALE
#define B1T 0x01f2 // 0.0304 * 2^LUX_SCALE
#define M1T 0x01be // 0.0272 * 2^LUX_SCALE

#define K2T 0x0080 // 0.250 * 2^RATIO_SCALE
#define B2T 0x0214 // 0.0325 * 2^LUX_SCALE
#define M2T 0x02d1 // 0.0440 * 2^LUX_SCALE

#define K3T 0x00c0 // 0.375 * 2^RATIO_SCALE
#define B3T 0x023f // 0.0351 * 2^LUX_SCALE
#define M3T 0x037b // 0.0544 * 2^LUX_SCALE

#define K4T 0x0100 // 0.50 * 2^RATIO_SCALE
#define B4T 0x0270 // 0.0381 * 2^LUX_SCALE
#define M4T 0x03fe // 0.0624 * 2^LUX_SCALE

#define K5T 0x0138 // 0.61 * 2^RATIO_SCALE
#define B5T 0x016f // 0.0224 * 2^LUX_SCALE
#define M5T 0x01fc // 0.0310 * 2^LUX_SCALE

#define K6T 0x019a // 0.80 * 2^RATIO_SCALE
#define B6T 0x00d2 // 0.0128 * 2^LUX_SCALE
#define M6T 0x00fb // 0.0153 * 2^LUX_SCALE

#define K7T 0x029a // 1.3 * 2^RATIO_SCALE
#define B7T 0x0018 // 0.00146 * 2^LUX_SCALE
#define M7T 0x0012 // 0.00112 * 2^LUX_SCALE

#define K8T 0x029a // 1.3 * 2^RATIO_SCALE
#define B8T 0x0000 // 0.000 * 2^LUX_SCALE
#define M8T 0x0000 // 0.000 * 2^LUX_SCALE

// lux equation approximation without floating point calculations
//////////////////////////////////////////////////////////////////////////////
// Routine: unsigned int CalculateLux(unsigned int ch0, unsigned int ch0)
//
// Description: Calculate the approximate illuminance (lux) given the raw
// channel values of the TSL2560. The equation if implemented
// as a piece-wise linear approximation.
//
// Arguments: unsigned int iGain - gain, where 0:1X, 1:16X
// unsigned int tInt - integration time, where 0:13.7mS, 1:100mS, 2:402mS, 3:Manual
// unsigned int ch0 - raw channel value from channel 0 of TSL2560
// unsigned int ch1 - raw channel value from channel 1 of TSL2560
//
// Return: unsigned int - the approximate illuminance (lux)
//
//////////////////////////////////////////////////////////////////////////////
static WORD CalculateLux (BYTE iGain, BYTE tInt, WORD ch0, WORD ch1)
{
	unsigned long	chScale;
	unsigned long	channel1;
	unsigned long	channel0;
	
	unsigned long	ratio1;
	unsigned long	ratio;
	unsigned int	b, m;
	unsigned long	temp;
	unsigned long	lux;
	
	//------------------------------------------------------------------------
	//	no processing for MANUAL integration time
	if (tInt == TIM_MANUAL)
		return ch0;

	//------------------------------------------------------------------------
	//	check for overflow of either channel
	if ((ch0 == 0xFFFF) || (ch1 == 0xFFFF))
		return 0xFFFF;

	//------------------------------------------------------------------------
	// first, scale the channel values depending on the gain and integration time
	// 16X, 402mS is nominal.
	// scale if integration time is NOT 402 msec
	switch (tInt)
	{
		case 0: // 13.7 msec
			chScale = CHSCALE_TINT0;
			break;
		case 1: // 101 msec
			chScale = CHSCALE_TINT1;
			break;
		case 2: // assume no scaling
			chScale = (1 << CH_SCALE);
			break;
	}
	
	// scale if gain is NOT 16X
	if (!iGain) 
		chScale = chScale << 4; // scale 1X to 16X
	// scale the channel values
	channel0 = (ch0 * chScale) >> CH_SCALE;
	channel1 = (ch1 * chScale) >> CH_SCALE;
	
	//------------------------------------------------------------------------
	// find the ratio of the channel values (Channel1/Channel0)
	// protect against divide by zero
	ratio1 = 0;
	if (channel0 != 0) 
		ratio1 = (channel1 << (RATIO_SCALE+1)) / channel0;

	// round the ratio value
	ratio = (ratio1 + 1) >> 1;

	// is ratio <= eachBreak ?
	if ((ratio >= 0) && (ratio <= K1T))
		{b=B1T; m=M1T;}
	else if (ratio <= K2T)
		{b=B2T; m=M2T;}
	else if (ratio <= K3T)
		{b=B3T; m=M3T;}
	else if (ratio <= K4T)
		{b=B4T; m=M4T;}
	else if (ratio <= K5T)
		{b=B5T; m=M5T;}
	else if (ratio <= K6T)
		{b=B6T; m=M6T;}
	else if (ratio <= K7T)
		{b=B7T; m=M7T;}
	else if (ratio > K8T)
		{b=B8T; m=M8T;}
	else
		{}
	
	temp = ((channel0 * b) - (channel1 * m));
	
	// do not allow negative lux value
	if (temp < 0) 
		temp = 0;

	// round lsb (2^(LUX_SCALE-1))
	temp += (1 << (LUX_SCALE-1));
	// strip off fractional portion
	lux = temp >> LUX_SCALE;

	return lux;
}

//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
/* currently no debug function */
}